/***********************************************************************/
/*                                                                     */
/*       (c) Copyright 2010                                           */
/*       All rights reserved                                           */
/*                                                                     */
/*       Programs written by Michael Bao                            */
/*       Department of Computer Science                                */
/*       New Jersey Institute of Technology                            */
/*       University Heights, Newark, NJ 07102, USA                     */
/*       In the group of Jason T. L. Wang (NJIT)                */
/*                                                                     */
/*	 Permission to use, copy, modify, and distribute this	       */
/*	 software and its documentation for any purpose and without    */
/*	 fee is hereby granted, provided that this copyright	       */
/*	 notice appears in all copies.   Programmer(s) makes no	       */
/*	 representations about the suitability of this		       */
/*	 software for any purpose.  It is provided "as is" without     */
/*	 express or implied warranty.				       */
/*                                                                     */
/***********************************************************************/

#include "Genome.h"
#define THRESHOLD 11.0
// complement of a nucleotide
char complement(char c)
{
  switch(c)
  {
    case 'U':
    case 'u':
    case 'T':
    case 't': return 'A';
    case 'G':
    case 'g': return 'C';
    case 'C':
    case 'c': return 'G';
    case 'A':
    case 'a': return 'T';
    default: return c;
  }
}
Genome::Genome() {
	// TODO Auto-generated constructor stub

}

Genome::~Genome() {
	// TODO Auto-generated destructor stub
}

int Genome::GetKnownRNA(string file, string genome, double& mean, double& stddev)
{
	ifstream RnaPosFile (file.c_str());
	vector<double> AllGC;
	vector<double> CG;
	vector<double> GG;
	vector<double> CC;
	vector<double> vGC;
	FreqR freq;
	string rna;
	string ncrna;
	int Index;
	string sBegin, sEnd;
	int iBegin, iEnd, iCount;
	double MeanGC, HighestGC, LowestGC;
	HighestGC = 0.0;
	LowestGC = 100.0;
	iCount = 0;
	Index = 0;
	iBegin = 0;
	iEnd = 0;
	TotalRNA = 0;
	MeanGC = 0.0;
	SeqR.GCContM = 0;
	SeqR.GCContS = 0;
	SeqR.CGSeqM = 0;
	SeqR.CGSeqS = 0;
	SeqR.GGSeqM = 0;
	SeqR.GGSeqS = 0;
	SeqR.CCSeqM = 0;
	SeqR.CCSeqS = 0;
	if (RnaPosFile.is_open())
	{
		double Sum = 0;
		while(!RnaPosFile.eof())
		{
			RNASeq seq;
			RNASeq rand;
			double GC;
			pair<int ,int > Pos;
			getline(RnaPosFile, rna);
			if (rna.size() < 3)
				continue;
			Index = rna.find("\t");
			sBegin = rna.substr(0, Index);
			sEnd = rna.substr(Index + 1);
			iBegin = atoi(sBegin.c_str());
			iEnd = atoi(sEnd.c_str());
			Pos.first = iBegin;
			Pos.second = iEnd;
			m_KnownRNAPos.push_back(Pos);

			ncrna = genome.substr(iBegin, iEnd - iBegin);
			m_KnownRNA.push_back(ncrna);
			seq.init_seq(ncrna, "Part", 0, "", "");
			rand.init_seq(RandomSequence(ncrna.size(), GenomeGC), "Part", 0, "", "");
			GC = seq.calcGCContent();
			freq = rand.calcSeqVar();
			MeanGC += GC;

			SeqR.GCContM += freq.GCContent;
			SeqR.CGSeqM += freq.CGSeq;
			SeqR.CCSeqM += freq.CCSeq;
			SeqR.GGSeqM += freq.GGSeq;
			iCount++;
			AllGC.push_back(GC);
			CG.push_back(freq.CGSeq);
			GG.push_back(freq.GGSeq);
			CC.push_back(freq.CCSeq);
			vGC.push_back(freq.GCContent);

			if (GC > HighestGC)
				HighestGC = GC;

			if (GC < LowestGC)
				LowestGC = GC;
			TotalRNA++;
		}

		MeanGC /= iCount;
		SeqR.GCContM /= iCount;
		SeqR.CGSeqM /= iCount;
		SeqR.CCSeqM /= iCount;
		SeqR.GGSeqM /= iCount;

		// calculate standard deviation
		for (int i = 0; i < iCount; i++)
		{
			Sum += pow(AllGC.at(i) - MeanGC, 2);
			SeqR.CGSeqS += pow(CG.at(i) - SeqR.CGSeqM, 2);
			SeqR.CCSeqS += pow(CC.at(i) - SeqR.CCSeqM, 2);
			SeqR.GGSeqS += pow(GG.at(i) - SeqR.CGSeqM, 2);
			SeqR.GCContS += pow(vGC.at(i) - SeqR.GCContM, 2);
		}

		mean = MeanGC;
		stddev = sqrt(Sum / (iCount - 1));
		SeqR.CGSeqS = sqrt(SeqR.CGSeqS / (iCount -1));
		SeqR.CCSeqS = sqrt(SeqR.CCSeqS / (iCount -1));
		SeqR.GGSeqS = sqrt(SeqR.GGSeqS / (iCount -1));
		SeqR.GCContS = sqrt(SeqR.GCContS / (iCount - 1));
	}
	m_FN = new int[m_KnownRNA.size()];
	for (int i = 0 ; i < m_KnownRNA.size(); i++)
	{
		m_FN[i] = m_KnownRNA.at(i).size();
		for (int j = 0; j < m_KnownRNA.at(i).size() - K_SIZE; j++)
		{
			string str = m_KnownRNA.at(i).substr(j, K_SIZE);
			RNAFreq[H((char*)str.c_str())]++;
		}
	}
	return iCount;
}
// retain smyRNA's hashing function
// the hashing function for 5-substring
inline int Genome::H(char* string)
{
  int i;
  int k=0;
  for(i=0; i<K_SIZE; i++)
  {
    switch(string[i])
    {
      case 'A': k +=0; break;
      case 'C': k +=1; break;
      case 'G': k +=2; break;
      case 'U':
      case 'T': k +=3; break;
      default:
    	  return -1;
    }
    k<<=2;
  }
  return k>>2;
}

void Genome::GetGenomeFromFile(string file, float& ATGCRatio)
{
	ifstream GenomeFile;
	string line;
	char* genomeseq;
	float AT = 0;
	float GC = 0;
	vector<string> ncrna;
	string temp;
	int pos = 0;
	char nucleotide;
	double ratio;

	GenomeFile.open(file.c_str(), ios::in);
	m_genome = "";
	if (GenomeFile.is_open())
	{
		GenomeFile.seekg(0, ios_base::end);
		GenomeSize = GenomeFile.tellg();
		GenomeFile.seekg(0, ios_base::beg);
	//	GenomeSize *= 2;
		data = new char[GenomeSize];
		while (!GenomeFile.eof())
		{
			nucleotide = GenomeFile.get();
			if (nucleotide == '\n')
			{
				GenomeSize--;
				continue;
			}
			else if (nucleotide != 'G' && nucleotide != 'C' &&
					nucleotide != 'T' && nucleotide != 'A' )
			{
				GenomeSize--;
				continue;
			}
			else if ((int)nucleotide == 0)
			{

				GenomeSize--;
				continue;
			}
			data[pos++] = nucleotide;
		}
		GenomeSize = pos;
		for (int i = 0; i < pos; i++)
		{
			nucleotide = data[i];
		//	data[GenomeSize - i -1] = complement(nucleotide);

			if (nucleotide == 'G' || nucleotide == 'C')
				GC++;
		}
		m_genome = string(data);
		GenomeGC = GC / GenomeSize;

	}

}

// retain smyRNA's hashing function
// the hashing function for 5-substring
int H(char* seq)
{
  int i;
  int k=0;
  string str(seq);
  for(i=0; i< str.length(); i++)
  {

    switch(seq[i])
    {
      case 'A':
      case 'a': k +=0; break;
      case 'C':
      case 'c': k +=1; break;
      case 'G':
      case 'g': k +=2; break;
      case 'U':
      case 'u':
      case 't':
      case 'T': k +=3; break;
      default:
        return -1;
    }

    k<<=2;
 }
  return k>>2;
}

void Genome::Initialize(string file, bool ncRNA, string ncRNAfile)
{
	vector<string> strKnownRNA;
	double GenomeSum = 0.0;
	double ratio;
	double RNASum = 0.0;
	// GENERATE FREQ TABLES
	GenomeFreq = new int[(int)K_POSSIB];
	RNAFreq = new int[(int)K_POSSIB];
	for (int i = 0; i < K_POSSIB; i++)
	{
		GenomeFreq[i] = 1;
		RNAFreq[i] = 1;
	}
	GetGenomeFromFile(file, ATGCRatio);
	hasNCRNA = ncRNA;

	TotalRNA = GetKnownRNA(ncRNAfile, m_genome, KnownRNAMean, KnownRNAStdDev);
	ifstream MeanMod (MODEL_FILE);
	ifstream SdMod (STDDEV_MODEL_FILE);
	if (MeanMod.fail() || SdMod.fail())
	{
		vector<LinearPoint> AllMeans;
		vector<LinearPoint> AllSd;

		for (int i = 1; i <= TOTAL_SAMPLES; i++)
		{
			int iSize = i * 50;
			double Mean;
			double Sd;
			vector<float> AllData;
			LinearPoint MeanPoint;
			LinearPoint SdPoint;
			MeanPoint.x = iSize;
			SdPoint.x = iSize;
			for (int y = 0; y < SAMPLE_SIZE; y++)
			{
				RNASeq seq;
				string structure;
				seq.init_seq(RandomSequence(iSize, GenomeGC), "Test", 0, "", "");
				AllData.push_back(seq.calcMFE(structure));
			}
			GetDistrib(AllData, Mean, Sd);
			MeanPoint.y = Mean;
			SdPoint.y = Sd;
			AllMeans.push_back(MeanPoint);
			AllSd.push_back(SdPoint);
		}

		MeanModel.InitModel(AllMeans, false, string(MODEL_FILE));
		SdModel.InitModel(AllSd, false, string(STDDEV_MODEL_FILE));
	}
	else
	{
		MeanModel.InitModel(MODEL_FILE);
		SdModel.InitModel(STDDEV_MODEL_FILE);
	}

	for (int i = 1; i < 26; i++)
	{
		//printf("TEST MODEL SIZE %d:::Mean %f:::SD %f\n", i * 50, MeanModel.GetValue(i*50),
		//		SdModel.GetValue(i*50));
	}

	for (int i = 0; i < GenomeSize - K_SIZE; i++)
	{
		GenomeFreq[H(data+i)]++;
	}

	for (int i = 0; i < K_POSSIB; i++)
	{
		GenomeSum += GenomeFreq[i];
		RNASum += RNAFreq[i];
	}
	ratio = (double)GenomeSum / RNASum;
	logprobs = new double[(int)K_POSSIB];
	ofstream LogOutput;
	LogOutput.open("LogOutputNCRNASCOUT.txt", ios::out | ios::trunc);

	for (int i = 0; i < K_POSSIB; i++)
	{
		logprobs[i] = log10((double)(RNAFreq[i]*ratio)  / GenomeFreq[i]);
		LogOutput << logprobs[i] << endl;
	}
	LogOutput.close();
	AllowToStart = true;
}

void Genome::GetDistrib(vector<float> Input, double& mean, double& stddev)
{
	double sum = 0;
	for (int i = 0; i < Input.size(); i++)
	{
		sum += Input.at(i);
	}

	mean = sum / Input.size();
	sum = 0;
	for (int i = 0; i < Input.size(); i++)
	{
		sum += pow(Input.at(i) - mean, 2);
	}

	sum /= (Input.size() - 1);
	stddev = sqrt(sum);
}


int Genome::StartSearch()
{
	if (!AllowToStart || !hasNCRNA)
		return -1;

	students_t tdistrib(TotalRNA - 1);
	students_t mfedistrib(SAMPLE_SIZE-1);
	vector<float> ncRNAValues;
	//ofstream PPVSens ("PPVSens.txt");
	double GC;
	float MFE;
	double Probability;
	string structure;
	string rna;
	double* scoretable;
	int* maxl;
	int orig_index = 0;
	int index = WINDOW_SIZE;
	string search;
	bool found = false;
	int count = 0;
	int TP;
	int FN;
	int FP;
	double HighestPPV = 0.0;
	double HighestSense = 0.0;
	offset = 0;

	FastaOutput.open(FASTA_OUT, ios::out | ios::trunc);
	RNASeq seq;
	seq.LoadModel();
//	for (int i = 3; i <= 9; i++)
//	{
//	for (double l = -11.0; l < 0; l += 0.5)
//	{
//		for (double t = 4.0; t <= 11.0; t += 1.0)
//		{
	char buff[30] = OUTPUT_FILE;
	sprintf(buff, "%s", OUTPUT_FILE);
	Output.open(buff, ios::out | ios::trunc);
	TP = 0; FN = 0; FP = 0;
	// identify regions in genome
	scoretable = new double[GenomeSize];
	spots = new NCSpot[GenomeSize];
	maxl = new int[GenomeSize];
	spotcount = 0;
	maxl[0] = 0;
	scoretable[0] = 0.0;
	for(int i=1; i< GenomeSize-K_SIZE; i++)
	{
		if(scoretable[i-1] >= -3.5)
		{
			scoretable[i] = scoretable[i-1] + logprobs[H(data+i)];
			maxl[i] = maxl[i-1];

			if(scoretable[i] > spots[spotcount].score)
			{
				spots[spotcount].k= i;
				spots[spotcount].l= maxl[i];
				spots[spotcount].score = scoretable[i];
			//    printf("%f\n",spots[spotcount].score);
			}
		}
		else
		{
			if(spots[spotcount].k - spots[spotcount].l > 0)
				spotcount++;
			else
				spots[spotcount].score = 0.0;

			scoretable[i] = logprobs[H(data+i)];
			maxl[i] = i;
		}
	}

	for (int i = 0; i < spotcount; i++)
	{
		if (spots[i].score > 6.0 && spots[i].k - spots[i].l < 15000)
		{
			RNASeq newSeq;
			double seqprob;
			std::string newstringhere(m_genome.substr(spots[i].l, spots[i].k - spots[i].l));
			float MFE;
			newSeq.init_seq(newstringhere, "Test", 0, "", "");
			MFE = newSeq.calcMFE(structure);
			MFE = (MFE - MeanModel.GetValue(newstringhere.size())) / ( SdModel.GetValue(newstringhere.size() / sqrt(SAMPLE_SIZE)));
			newSeq.setMFE(MFE);
			newSeq.SetSeqVar(SeqR, TotalRNA);
			seqprob = newSeq.CalculateSeqProbability();
			if (newSeq.GetClassification() == 1.0)
				Output << spots[i].l << "\t" << spots[i].k << endl;
		}
	}
	delete [] scoretable;
	delete [] spots;
	delete [] maxl;

	Output.close();

//		} // threshold loop
//	} // lower value loop


	return count;

}

bool Genome::determineOverlapWithNcRna(int l, int k, int i)
{
	// loops through ncRNA
	if (l <= m_KnownRNAPos.at(i).second & l >= m_KnownRNAPos.at(i).first
		|| k <= m_KnownRNAPos.at(i).second && k >= m_KnownRNAPos.at(i).first )
		return true;
	else
		return false;
}

void Genome::determinePosNeg(int i, int l, int k, int &TP, int &FP, int &FN, bool &spillOver)
{


}

float Genome::GetProbability(float Value, students_t distrib, float mean, float sd)
{
	float tstat = (Value - mean) / sd;
	float prob = 1;

	prob = cdf(distrib, tstat);


	return prob;
}

float SumOfFloatVector(vector<float> Vector)
{
	float Sum = 0;
	for (int i = 0; i < Vector.size(); i++)
	{
		Sum += Vector.at(i);
	}
	return Sum;
}


void Genome::CreateTDistributionRand( float& mean, float& sd, int length, double GC)
{
	vector<float> AllMFE;
	double variance = 0.0;
	for (int i = 0; i < SAMPLE_SIZE; i++)
	{
		// create random ncRNA
		RNASeq seq;
		string temp;
		seq.init_seq(RandomSequence(length, GC), "Random", i, "", "");
		AllMFE.push_back(seq.calcMFE(temp));
	}
	mean = SumOfFloatVector(AllMFE) / AllMFE.size();
	for (int i = 0; i < AllMFE.size(); i++)
	{
		variance += pow((AllMFE.at(i) - mean), 2);
	}
	sd = sqrt(variance / (SAMPLE_SIZE - 1));
}

string Genome::RandomSequence(int length, double GC)
{
	// generate a random sequence with specified length and GC content
	string sequence = "";
	int random = 0;
	srand( time( NULL ) );
	for (int i = 0; i < length; i++)
	{
		random = rand() % 100 + 1;
		if (random <= GC * 100)
		{
			random = rand() % 100 + 1;
			if (random <= 50)
				sequence.append("G");
			else
				sequence.append("C");
		}
		else
		{
			random = rand() % 100 + 1;
			if (random <= 50)
				sequence.append("A");
			else
				sequence.append("T");
		}
	}
	return sequence;
}
