/***********************************************************************/
/*                                                                     */
/*       (c) Copyright 2010                                           */
/*       All rights reserved                                           */
/*                                                                     */
/*       Programs written by Michael Bao                            */
/*       Department of Computer Science                                */
/*       New Jersey Institute of Technology                            */
/*       University Heights, Newark, NJ 07102, USA                     */
/*       In the group of Jason T. L. Wang (NJIT)                */
/*                                                                     */
/*	 Permission to use, copy, modify, and distribute this	       */
/*	 software and its documentation for any purpose and without    */
/*	 fee is hereby granted, provided that this copyright	       */
/*	 notice appears in all copies.   Programmer(s) makes no	       */
/*	 representations about the suitability of this		       */
/*	 software for any purpose.  It is provided "as is" without     */
/*	 express or implied warranty.				       */
/*                                                                     */
/***********************************************************************/

#ifndef GENOME_H_
#define GENOME_H_

#define WINDOW_SIZE 30
#define INCREMENT_SIZE 10
#define SAMPLE_SIZE 250
#define ALPHA_MFE 0.05
#define ALPHA_GC 0.30
#define MODEL_FILE "MeanRegression.model"
#define STDDEV_MODEL_FILE "StdDevRegression.model"
#define TOTAL_SAMPLES 15
#define REMOVE_SIZE 2

#include <ctype.h>
#include "stdio.h"
#include "stdlib.h"
#include <iostream>
#include "RNASeq.h"
#include "LinReg.h"
#include <vector>
#include <time.h>
#include <fstream>
#include <string>
#include "SVMRegression.h"
#include <exception>
#include "math.h"
//#include <uninstd.h>
#define OUTPUT_FILE "results.txt"
#define FASTA_OUT	"results.fasta"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

using namespace std;

struct Position
{
	int Start;
	int End;
};

struct NCSpot
{
	int l;
	int k;
	double score;
};


class Genome {
public:
	Genome();
	virtual ~Genome();

	void Initialize(string, bool, string);
	int StartSearch();

	void GetGenomeFromFile(string file, float& ATGCRatio);

	int GetKnownRNA(string file, string genome, double&, double&);


private:
	string m_genome;
	vector<string> m_KnownRNA;
	vector<pair <int, int> > m_KnownRNAPos;
	int* m_FN;
	vector< pair<int, int> > ORF;
	Position pFound; // temporary storage for ncRNA regions
	ofstream Output;
	ofstream FastaOutput;

	double KnownRNAMean;
	double KnownRNAStdDev;
	int TotalRNA;

	int GenomeSize;
	float ATGCRatio;
	double GenomeGC;

	int* GenomeFreq;
	int* RNAFreq;
	int H(char* string);

	void CreateTDistributionRand(float&, float&, int, double); // random sequences
	float GetProbability(float, students_t, float, float);
	void GetDistrib(vector<float>, double&, double&);
	string RandomSequence(int, double);
	double CalculateSeqProbability(RNASeq);
	bool TickFind();
	bool determineOverlapWithNcRna(int l, int k, int i);
	void determinePosNeg(int i, int l, int k, int &TP, int &FP, int &FN, bool &spillOver);

	LinReg MeanModel;
	LinReg SdModel;

	bool AllowToStart;
	bool hasNCRNA;

	int offset;

	// mean and std dev for sequence variables
	SeqRVar SeqR;

	char* data;
	NCSpot* spots;
	int spotcount;

	double* logprobs;

};

#endif /* GENOME_H_ */
