/***********************************************************************/
/*                                                                     */
/*       (c) Copyright 2010                                           */
/*       All rights reserved                                           */
/*                                                                     */
/*       Programs written by Michael Bao                            */
/*       Department of Computer Science                                */
/*       New Jersey Institute of Technology                            */
/*       University Heights, Newark, NJ 07102, USA                     */
/*       In the group of Jason T. L. Wang (NJIT)                */
/*                                                                     */
/*	 Permission to use, copy, modify, and distribute this	       */
/*	 software and its documentation for any purpose and without    */
/*	 fee is hereby granted, provided that this copyright	       */
/*	 notice appears in all copies.   Programmer(s) makes no	       */
/*	 representations about the suitability of this		       */
/*	 software for any purpose.  It is provided "as is" without     */
/*	 express or implied warranty.				       */
/*                                                                     */
/***********************************************************************/

#include "KmerProb.h"
#include <sys/time.h>
#include "math.h"

#define MAX_SIZE 4639675

bool IdentifyNucleotide(char c)
{
	switch(c)
	{
	case 'A':
	case 'a':
	case 'T':
	case 't':
	case 'C':
	case 'c':
	case 'G':
	case 'g':
	case 'U':
	case 'u':
		return true;
	default:
		return false;

	}
}



KmerProb::KmerProb(string input, string genome, string KnownRNA, bool over) {
	RNASequence=input;
	strGenomeFile = genome;
	strKnownRnaFile = KnownRNA;
	GenomeSequence = genome;
	bGotGenome = false;
	bOverride = over;
	bGotFrequencies = false;
}

KmerProb::~KmerProb() {
	// TODO Auto-generated destructor stub
}

// retain smyRNA's hashing function
// the hashing function for 5-substring
inline int KmerProb::H(char* string)
{
  int i;
  int k=0;
  for(i=0; i<K_SIZE; i++)
  {
    switch(string[i])
    {
      case 'A': k +=0; break;
      case 'C': k +=1; break;
      case 'G': k +=2; break;
      case 'U':
      case 'T': k +=3; break;
      default:
    	  return -1;
    }
    k<<=2;
  }
  return k>>2;
}

/*
 * read in genome and known RNA position files
 */
void KmerProb::open()
{
	string buffer;
	int GenomeFreqStart = 0;
	int GenomeMod = 0;
	string seq;
	char* genomeseq;
	GenomeSize = GenomeSequence.length();
	printf("Opening genome..\n");
	genomeseq = new char[GenomeSize];
	for(int i = 0; i < GenomeSize; i++)
	{
		if (IdentifyNucleotide(GenomeSequence[i]))
		{
			genomeseq[i] = GenomeSequence[i];
			if (i >= (int)RNASequence.length() && i < GenomeSize - GenomeMod - RNASequence.length() - 29)
			{
				if (i == RNASequence.length())
					printf("Sequence Length: %d\n", RNASequence.length());
				seq = GenomeSequence.substr(i, RNASequence.length());
				if (id == 3 && seq.size() > RNASequence.length())
					printf(">>>MESSED UP READ. LENGTH: %d, Genome Freq Start: %d\n", seq.size(), GenomeFreqStart);
				seq.resize(RNASequence.length());
				int hash = H((char*)seq.c_str());
				if (i % 50000 == 0)
					printf("%d - \tNum: %d\n", i, hash);
				if (GenomeFrequency.find(hash) == GenomeFrequency.end())
					GenomeFrequency[hash] = 1;
				else
					GenomeFrequency[hash]++;
				GenomeFreqStart++;
			}
		}
		else
		{
			GenomeMod++;
			printf("Unknown Nucleotide at position %d: %c\n", i, GenomeSequence[i]);
		}
		// calculate genome hashes as we go through

		if (i== MAX_SIZE)
		{
			printf("Genome too large going with first %d nucleotides\n", MAX_SIZE);
			break;
		}
	}

	delete [] genomeseq;

	KnownRnaFile.open(strKnownRnaFile.c_str(), ios::in);
	if (KnownRnaFile.is_open())
	{

		string begin;
		string end;
		subsequence subseq;
		while (!KnownRnaFile.eof() )
		{
			getline(KnownRnaFile, buffer);
			// get RNA Positions from file
			begin = buffer.substr(0, buffer.find("\t"));
			end = buffer.substr(buffer.find("\t") + 1);
			subseq.begin = atoi(begin.c_str());
			subseq.end = atoi(end.c_str());
			subseq.name = "Default";
			KnownNcRna.push_back(subseq);
		}
	}
	KnownRnaFile.close();

}

void KmerProb::CalculateKnownNcRnaHash()
{
	string sequence;

	GenomeKmerCount = 0;

	printf("Start calculating... \n");

	for(int i = 0; i < KnownNcRna.size(); i++)
	{
		// iterate through vector and retrieve the appropriate string and calculate hash
		sequence = GenomeSequence.substr(KnownNcRna.at(i).begin, KnownNcRna.at(i).end - KnownNcRna.at(i).begin);
		KnownNcRna.at(i).hash=H((char*)sequence.c_str());
		if (RNAFrequency.find(KnownNcRna.at(i).hash) == RNAFrequency.end())
			RNAFrequency[KnownNcRna.at(i).hash]=1;
		else
			RNAFrequency[KnownNcRna.at(i).hash]++;
	}


	printf("Done with calculating...\n");
}

void KmerProb::InitiateFrequencies(map<int,int> G, map<int,int> N)
{
	bGotFrequencies = true;
	GenomeFrequency = G;
	RNAFrequency = N;
}

double KmerProb::getLogSingleSequence()
{
	// get the log probability of a single sequence
	int KmerSize;
	int KmerHash;
	int RNACount;
	int GenomeCount;
	int Count;
	double RNAFreq;
	double GenomeFreq;
	double LogProb;
	double ratio;
	vector<int> AllHash;
	map<int, int>::iterator it;
	string TempRNA;

	TempRNA = RNASequence;

	if (!bOverride && !bGotFrequencies)
		CalculateKnownNcRnaHash();


	KmerSize=K_SIZE;
	RNACount = 0;
	GenomeCount = 0;
	LogProb = 0.0;
	Count = 0;
	for (it = GenomeFrequency.begin(); it != GenomeFrequency.end(); it++)
		GenomeCount += it->second;

	for (it = RNAFrequency.begin(); it != RNAFrequency.end(); it++)
		RNACount += it->second;


	printf("RNA Count: %d\n", RNACount);
	printf("Genome Count: %d\n", GenomeCount);
	ratio = (double)GenomeCount / 1024;

	printf("%f\n", ratio);
	for (int i = 0; i< TempRNA.size() - KmerSize; i++)
	{
		KmerHash = H((char*)TempRNA.substr(i, KmerSize).c_str());
		double tee = log10( ( ratio * 1) / GenomeFrequency[KmerHash]);
		if (RNAFrequency[KmerHash] == 0)
			tee = log10( ratio / GenomeFrequency[KmerHash]);
	//	printf("%d %d %f\n", RNAFrequency[KmerHash], GenomeFrequency[KmerHash], tee);
		// get total lob probability score
		LogProb += tee;
	}

	// clean up
	GenomeFrequency.clear();
	RNAFrequency.clear();
	GenomeSequence = "";

	return LogProb;

}
