/***********************************************************************/
/*                                                                     */
/*       (c) Copyright 2010                                           */
/*       All rights reserved                                           */
/*                                                                     */
/*       Programs written by Michael Bao                            */
/*       Department of Computer Science                                */
/*       New Jersey Institute of Technology                            */
/*       University Heights, Newark, NJ 07102, USA                     */
/*       In the group of Jason T. L. Wang (NJIT)                */
/*                                                                     */
/*	 Permission to use, copy, modify, and distribute this	       */
/*	 software and its documentation for any purpose and without    */
/*	 fee is hereby granted, provided that this copyright	       */
/*	 notice appears in all copies.   Programmer(s) makes no	       */
/*	 representations about the suitability of this		       */
/*	 software for any purpose.  It is provided "as is" without     */
/*	 express or implied warranty.				       */
/*                                                                     */
/***********************************************************************/

#include "LinReg.h"
#include <cstdio>
#include <stdlib.h>

LinReg::LinReg() {
	// TODO Auto-generated constructor stub

}

LinReg::~LinReg() {
	// TODO Auto-generated destructor stub
}

void LinReg::InitModel(string file)
{
	string line;
	ifstream model (file.c_str(), ios::in);
	if (model.is_open())
	{
		while(!model.eof())
		{
			LinearPoint LP;
			int index;
			string first, second;
			getline(model, line);

			index = line.find('\t');

			if (index != string::npos)
			{
				first = line.substr(0, index);
				second = line.substr(index);

				LP.x = atof(first.c_str());
				LP.y = atof(second.c_str());

				Points.push_back(LP);
			}

		}
	}

	InitModel(Points);
}

void LinReg::InitModel(vector<LinearPoint> f, bool file, string output)
{
	Points = f;

	// get X and Y means
	float meanX, meanY;
	float num, denom;

	meanX = 0;
	meanY = 0;
	num = 0;
	denom = 0;

	for (int i = 0; i < Points.size(); i++)
	{
		meanX += Points.at(i).x;
		meanY += Points.at(i).y;
	}

	meanX /= Points.size();
	meanY /= Points.size();

	// calculate slope
	for (int i = 0; i < Points.size(); i++)
	{
		num += (Points.at(i).x - meanX) * (Points.at(i).y - meanY);
		denom += pow(Points.at(i).x - meanX, 2);
	}

	slope = num / denom;
	intercept = meanY - slope * meanX;

	if (!file)
	{
		ofstream out(output.c_str());
		for (int i = 0; i < Points.size(); i++)
		{
			out << Points.at(i).x << "\t" << Points.at(i).y << endl;
		}
		out.close();
	}
}

float LinReg::GetValue(float x)
{
	return x * slope + intercept;
}
