/***********************************************************************/
/*                                                                     */
/*       (c) Copyright 2010                                           */
/*       All rights reserved                                           */
/*                                                                     */
/*       Programs written by Michael Bao                            */
/*       Department of Computer Science                                */
/*       New Jersey Institute of Technology                            */
/*       University Heights, Newark, NJ 07102, USA                     */
/*       In the group of Jason T. L. Wang (NJIT)                */
/*                                                                     */
/*	 Permission to use, copy, modify, and distribute this	       */
/*	 software and its documentation for any purpose and without    */
/*	 fee is hereby granted, provided that this copyright	       */
/*	 notice appears in all copies.   Programmer(s) makes no	       */
/*	 representations about the suitability of this		       */
/*	 software for any purpose.  It is provided "as is" without     */
/*	 express or implied warranty.				       */
/*                                                                     */
/***********************************************************************/


#include "RNASeq.h"

RNASeq::RNASeq(void)
{
	deviation_model = new LinReg();
	deviation_model->InitModel(DEVIATION_MODEL);
}

RNASeq::~RNASeq(void)
{
}

int RNASeq::init_seq(string seq, string name, int id, string genome, string RNApos)
{
	m_id = id;
	m_sequence = seq;
	m_name = name;
	m_length = seq.length();
	m_genome=genome;
	m_RNApos=RNApos;
	model_file = svm_load_model("SVMTrainData.txt.model");
	bOverride = false;
	return 0;
}

void RNASeq::SetSeqVar(SeqRVar R, int count)
{
	SeqR = R;
	TotalRNA = count;
}

FreqR RNASeq::calcSeqVar()
{
	FreqR freq;
	int PrevMarker = -1;
	double C, G;
	int Marker = -1;
	freq.CCSeq = 0;
	freq.CGSeq = 0;
	freq.GCContent = 0;
	freq.GGSeq = 0;
	C = 0;
	G = 0;
	for (int i = 0 ; i < m_sequence.size(); i++)
	{
		PrevMarker = Marker;
		switch(m_sequence[i])
		{
			case 'C':
			{
				Marker = 0;
				C++;
				break;
			}
			case 'G':
			{
				Marker = 1;
				G++;
				break;
			}
		}

		if (i != 0)
		{
			switch (PrevMarker)
			{
			case 0:
			{
				switch (Marker)
				{
					case 0:
					{
						freq.CCSeq++;
						break;
					}
					case 1:
					{
						freq.CGSeq++;
						break;
					}
				}
				break;
			}
			case 1:
			{
				switch (Marker)
				{
					case 1:
					{
						freq.GGSeq++;
						break;
					}
				}
				break;
			}
			}

		}
	}

	freq.GCContent = (C + G) / m_sequence.size();
	freq.CCSeq /= (m_sequence.size() - 1);
	freq.CGSeq /= (m_sequence.size() - 1);
	freq.GGSeq /= (m_sequence.size() - 1);

//	printf("%f %f %f %f\n", freq.GCContent, freq.CCSeq, freq.CGSeq, freq.GGSeq);

	return freq;
}

double RNASeq::CalculateSeqProbability()
{
	FreqR freq;
	students_t tdistrib(TotalRNA - 1);
	double GCProb;
	double CGProb;
	double CCProb;
	double GGProb;
	freq = calcSeqVar();

	GCProb = GetProbability(freq.GCContent, tdistrib, SeqR.GCContM, SeqR.GCContS);
	CGProb = GetProbability(freq.CGSeq, tdistrib, SeqR.CGSeqM, SeqR.CGSeqS);
	CCProb = GetProbability(freq.CCSeq, tdistrib, SeqR.CCSeqM, SeqR.CCSeqS);
	GGProb = GetProbability(freq.GGSeq, tdistrib, SeqR.GGSeqM, SeqR.GGSeqS);

	SequenceAnalysisProbValue = GCProb * GC_WEIGHT + CGProb * CG_WEIGHT + CCProb * CC_WEIGHT + GGProb * GG_WEIGHT;
	return SequenceAnalysisProbValue;
}

float RNASeq::GetProbability(float Value, students_t distrib, float mean, float sd)
{
	float tstat = (Value - mean) / sd;
	float prob = 1;

	prob = cdf(distrib, tstat);


	return prob;
}

double RNASeq::calcGCContent()
{
	int count = 0;
	for (int i = 0 ; i < m_sequence.size(); i++)
	{
		if (m_sequence[i] == 'G' ||
				m_sequence[i] == 'g' || m_sequence[i] == 'C' ||
				m_sequence[i] == 'c' )
		{
			count++;
		}
	}

	return ((double)count/  m_sequence.size());
}

float RNASeq::calcMFE(string& out_str)
{
//	char* sequence;
	char* structure;

	// initalize
	Vienna::temperature = 30.;
	//printf("MFE SIZE: %d\n", m_sequence.size());
	//	printf("CALC MFE\n");
		// calculate MFE using Vienna RNA package
	structure = (char* ) Vienna::space(sizeof(char)*(m_sequence.size()+1));
	MFE = Vienna::fold(m_sequence.c_str(), structure);
	//printf("FINISH MFE\n");
	Vienna::free_arrays();

	return MFE;
}

double RNASeq::calcKmer()
{
	KmerProb *calcKmer;

	printf("CALC KMER\n");
	calcKmer = new KmerProb(m_sequence, m_genome, m_RNApos, bOverride);
	calcKmer->id = m_id;
	if (bOverride)
		calcKmer->InitiateFrequencies(GFreq, NFreq);
	else
		calcKmer->open();

	KmerLogProb=calcKmer->getLogSingleSequence();

	delete calcKmer;

	return KmerLogProb;
}

void RNASeq::OverrideKmer(map<int, int> G, map<int, int> N)
{
	bOverride = true;
	GFreq = G;
	NFreq = N;
}

double RNASeq::GetClassification()
{
	svm_node kmerNode;
	svm_node probNode;
	svm_node nullNode;
	// array of svm nodes
	svm_node nodes[3];

	kmerNode.index = 0;
	kmerNode.value = SequenceAnalysisProbValue;

	probNode.index = 1;
	probNode.value = MFE;

	nullNode.index = -1;
	nullNode.value = 0;

	nodes[0] = kmerNode;
	nodes[1] = probNode;
	nodes[2] = nullNode;

	predicted_value = svm_predict(model_file, nodes);
	svm_destroy_model(model_file);

	return predicted_value;
}
