/***********************************************************************/
/*                                                                     */
/*       (c) Copyright 2010                                           */
/*       All rights reserved                                           */
/*                                                                     */
/*       Programs written by Michael Bao                            */
/*       Department of Computer Science                                */
/*       New Jersey Institute of Technology                            */
/*       University Heights, Newark, NJ 07102, USA                     */
/*       In the group of Jason T. L. Wang (NJIT)                */
/*                                                                     */
/*	 Permission to use, copy, modify, and distribute this	       */
/*	 software and its documentation for any purpose and without    */
/*	 fee is hereby granted, provided that this copyright	       */
/*	 notice appears in all copies.   Programmer(s) makes no	       */
/*	 representations about the suitability of this		       */
/*	 software for any purpose.  It is provided "as is" without     */
/*	 express or implied warranty.				       */
/*                                                                     */
/***********************************************************************/

#pragma once

#include <iostream>
#include <vector>
#include <math.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
//#include "centroid.h"
#include "KmerProb.h"
#include "LinReg.h"
#include "SVMRegression.h"
#include "svm.h"
#include <boost/math/distributions/students_t.hpp>
using namespace boost::math;

#define DEVIATION_MODEL "DeviationRegression.model"
#define GC_WEIGHT	0.6
#define CG_WEIGHT	0.1
#define GG_WEIGHT	0.15
#define CC_WEIGHT	0.15


namespace Vienna {
	extern "C" {
		#include  "utils.h"
		#include  "fold_vars.h"
		#include  "fold.h"
	};
};

struct FreqR
{
	double GCContent;
	double CGSeq;
	double GGSeq;
	double CCSeq;
};

struct SeqRVar
{
	double GCContM, GCContS;
	double CGSeqM, CGSeqS;
	double GGSeqM, GGSeqS;
	double CCSeqM, CCSeqS;
};

using namespace std;

class RNASeq
{
public:
	RNASeq(void);
	~RNASeq(void);

	int init_seq(string, string, int, string, string);
	void OverrideKmer(map<int, int>, map<int, int>);
	int seq_length() { return m_length; }
	double calcGCContent();

	void setMFE(float M) { MFE = M; }

	float calcMFE(string&);
	double calcKmer();
	FreqR calcSeqVar();
	double CalculateSeqProbability();

	double GetClassification();

	string GetSequence() { return m_sequence; }

	void SetSeqVar(SeqRVar, int);

	void LoadModel() {  }

private:

	int m_id;
	int m_length; // length of sequence
	string m_sequence; // the actual sequence in string form
	string m_name;

	// smyRNA k-mer prob calculation
	string m_genome;
	string m_ORFS;
	string m_RNApos;
	float m_threshold;

	float MFE;
	double KmerLogProb;

	svm_model* model_file;
	LinReg* deviation_model;

	double predicted_value;

	bool bOverride;
	map<int, int> GFreq;
	map<int, int> NFreq;

	int TotalRNA;

	// novel
	double SequenceAnalysisProbValue;
	// mean and std dev for sequence variables
	SeqRVar SeqR;

	float GetProbability(float, students_t, float, float);

};
