/***********************************************************************/
/*                                                                     */
/*       (c) Copyright 2010                                           */
/*       All rights reserved                                           */
/*                                                                     */
/*       Programs written by Michael Bao                            */
/*       Department of Computer Science                                */
/*       New Jersey Institute of Technology                            */
/*       University Heights, Newark, NJ 07102, USA                     */
/*       In the group of Jason T. L. Wang (NJIT)                */
/*                                                                     */
/*	 Permission to use, copy, modify, and distribute this	       */
/*	 software and its documentation for any purpose and without    */
/*	 fee is hereby granted, provided that this copyright	       */
/*	 notice appears in all copies.   Programmer(s) makes no	       */
/*	 representations about the suitability of this		       */
/*	 software for any purpose.  It is provided "as is" without     */
/*	 express or implied warranty.				       */
/*                                                                     */
/***********************************************************************/

#include "SVMRegression.h"

SVMRegression::SVMRegression() {
	// TODO Auto-generated constructor stub

}

SVMRegression::~SVMRegression() {
	// TODO Auto-generated destructor stub
}

string SVMRegression::RandomSequence(int length, double GC)
{
	// generate a random sequence with specified length and GC content
	string sequence = "";
	int random = 0;
	srand( time( NULL ) );
	for (int i = 0; i < length; i++)
	{
		random = rand() % 100 + 1;
		if (random <= GC * 100)
		{
			random = rand() % 100 + 1;
			if (random <= 50)
				sequence.append("G");
			else
				sequence.append("C");
		}
		else
		{
			random = rand() % 100 + 1;
			if (random <= 50)
				sequence.append("A");
			else
				sequence.append("T");
		}
	}
	return sequence;
}


int SVMRegression::TrainSVM()
{
	ifstream mean(MODEL_FILE);
	ifstream sd(STDDEV_MODEL_FILE);
	ofstream out ("Debug.txt");

	if (!mean.fail() && !sd.fail())
	{
		MeanMod = svm_load_model(MODEL_FILE);
		SdMod = svm_load_model(STDDEV_MODEL_FILE);
		mean.close();
		sd.close();
		return 1;
	}


	svm_node** allnodes = new svm_node*[TOTAL_SAMPLES];
	double* predicted = new double[TOTAL_SAMPLES];
	double* sdpredict = new double[TOTAL_SAMPLES];
	for (int i = 0; i < TOTAL_SAMPLES; i++)
	{
		// generate random samples to get mean and SD of MFE to be able to predict
		vector<float> IndivMFE;
		int RNASize = (i+1) * 50;
		double MFE = 0.0;
		double SD = 0.0;
		svm_node node;
		allnodes[i] = new svm_node[1];
		for (int y = 0; y < SAMPLE_MFE_SIZE; y++)
		{
			RNASeq seq;
			string structure;
			double TempMFE;
			seq.init_seq(RandomSequence(RNASize, 0.5), "Test", 0, "", "");
			TempMFE = seq.calcMFE(structure);
			MFE += TempMFE;
			IndivMFE.push_back(TempMFE);
		}
		MFE /= SAMPLE_MFE_SIZE;
		predicted[i] = MFE;
		node.index = 1;
		node.value = i * 50;
		allnodes[i][0] = node;


		for (int y = 0; y < SAMPLE_MFE_SIZE; y++)
		{
			SD += pow(IndivMFE.at(y) - MFE, 2);
		}

		SD /= SAMPLE_MFE_SIZE - 1;
		SD = sqrt(SD);
		sdpredict[i] = SD;

		out << "SET " << i << " - MFE: " << MFE << " SD: " << SD << endl;
	}

	out << "FINISH GETTING DATA" << endl;

	svm_problem prob;
	svm_parameter param;
	vector<double> CourseTestC;
	vector<double> CourseTestGamma;
	vector<double> FineTestC;
	vector<double> FineTestGamma;
	for(float i = pow(2.0f, -7); i <= pow(2.0f,3); i *= 4)
	{
		CourseTestC.push_back(i);
		CourseTestGamma.push_back(i);
	}
	prob.l = TOTAL_SAMPLES;
	prob.y = predicted;
	prob.x = allnodes;
	param.svm_type = EPSILON_SVR;
	param.kernel_type = RBF;
	param.cache_size = 25;
	param.eps = 0.001;
	param.p = pow(1.0f, -3);
	param.shrinking = 1;
	param.probability = 0;

	// determine C and gamma
	// use grid search; course

	// THE FOLLOWING IS TO CREATE MODEL FOR THE MEAN PREDICTION
	out << "START MEAN MODEL" << endl;
	int IndexCourseC = 0;
	int IndexCourseY = 0;
	int IndexFineC = 0;
	int IndexFineY = 0;
	double HighestCrossValidation = 0;
	double CV = 0;
	out << "COURSE CHECK" << endl;
	for (int c = 0; c < CourseTestC.size(); c++)
	{
		for (int y = 0; y < CourseTestGamma.size(); y++)
		{
			param.C = CourseTestC.at(c);
			param.gamma = CourseTestGamma.at(y);
			//model = svm_train(&prob, &param);
			CV = CrossValidation(prob, param);
			if (CV > HighestCrossValidation)
			{
				HighestCrossValidation = CV;
				IndexCourseC = c;
				IndexCourseY = y;
			}
		}
	}

	HighestCrossValidation = 0;
	CV = 0;

	// create fine values
	for(float i = CourseTestC.at(IndexCourseC); i <= CourseTestC.at(IndexCourseC+1); i *= pow(2, .25))
	{
		FineTestC.push_back(i);
	}

	for(float i = CourseTestGamma.at(IndexCourseY); i <= CourseTestGamma.at(IndexCourseY+1); i *= pow(2, .25))
	{
		FineTestGamma.push_back(i);
	}

	out << "FINE CHECK" << endl;
	// use grid search; fine
	for (int c = 0; c < FineTestC.size(); c++)
	{
		for (int y = 0; y < FineTestGamma.size(); y++)
		{
			param.C = FineTestC.at(c);
			param.gamma = FineTestGamma.at(y);
			CV = CrossValidation(prob, param);
			if (CV > HighestCrossValidation)
			{
				HighestCrossValidation = CV;
				IndexFineC = c;
				IndexFineY = y;
			}
		}
	}

	param.gamma = FineTestGamma.at(IndexFineY);
	param.C = FineTestC.at(IndexFineC);
	MeanMod = svm_train(&prob, &param);
	out << "SAVING MODEL" << endl;
	int ret = svm_save_model(MODEL_FILE, MeanMod);

	out << "DESTROY MEAN MODEL" << endl;
	svm_destroy_model(MeanMod);

	// THE FOLLOWING IS TO CREATE THE MODEL FOR STANDARD DEVIATION PREDICTION
	out << "START STANDARD DEVIATION MODEL" << endl;
	prob.y = sdpredict;
	IndexCourseC = 0;
	IndexCourseY = 0;
	IndexFineC = 0;
	IndexFineY = 0;
	HighestCrossValidation = 0;
	CV = 0;
	out << "COURSE CHECK" << endl;
	for (int c = 0; c < CourseTestC.size(); c++)
	{
		for (int y = 0; y < CourseTestGamma.size(); y++)
		{
			param.C = CourseTestC.at(c);
			param.gamma = CourseTestGamma.at(y);
			//model = svm_train(&prob, &param);
			CV = CrossValidation(prob, param);
			if (CV > HighestCrossValidation)
			{
				HighestCrossValidation = CV;
				IndexCourseC = c;
				IndexCourseY = y;
			}
		}
	}

	HighestCrossValidation = 0;
	CV = 0;

	FineTestC.clear();
	FineTestGamma.clear();
	// create fine values
	for(float i = CourseTestC.at(IndexCourseC); i <= CourseTestC.at(IndexCourseC+1); i *= pow(2, .25))
	{
		FineTestC.push_back(i);
	}

	for(float i = CourseTestGamma.at(IndexCourseY); i <= CourseTestGamma.at(IndexCourseY+1); i *= pow(2, .25))
	{
		FineTestGamma.push_back(i);
	}

	out << "FINE CHECK" << endl;
	// use grid search; fine
	for (int c = 0; c < FineTestC.size(); c++)
	{
		for (int y = 0; y < FineTestGamma.size(); y++)
		{
			param.C = FineTestC.at(c);
			param.gamma = FineTestGamma.at(y);
			CV = CrossValidation(prob, param);
			if (CV > HighestCrossValidation)
			{
				HighestCrossValidation = CV;
				IndexFineC = c;
				IndexFineY = y;
			}
		}
	}

	param.gamma = FineTestGamma.at(IndexFineY);
	param.C = FineTestC.at(IndexFineC);
	SdMod = svm_train(&prob, &param);
	out << "SAVE SD MODEL" << endl;
	ret = svm_save_model(STDDEV_MODEL_FILE, SdMod);
	out << "DESTROY MODEL" << endl;
	svm_destroy_model(SdMod);

	delete [] predicted;
	delete [] sdpredict;

	out << "RETURN" << endl;
	out.close();

	return ret;


}

double SVMRegression::CrossValidation(svm_problem& prob, svm_parameter& param)
{
	double rsq = 0;
	double *target = new double[prob.l];
	double total_error = 0;
	int total_correct = 0;
	double sumv = 0, sumy = 0, sumvv = 0, sumyy = 0, sumvy = 0;
	svm_cross_validation(&prob, &param, 5, target);
	for(int i = 0; i < prob.l; i++)
	{
		double y = prob.y[i];
		double v = target[i];
		total_error += (v-y)*(v-y);
		sumv += v;
		sumy+= y;
		sumvv += v*v;
		sumyy += y*y;
		sumvy += v* y;
	}

	rsq = ((prob.l*sumvy-sumv*sumy)*(prob.l*sumvy-sumv*sumy))/
			((prob.l*sumvv-sumv*sumv)*(prob.l*sumyy-sumy*sumy));

	return rsq;
}

double SVMRegression::PredictMean(int size)
{
	svm_node node;
	node.index = 1;
	node.value = size;
	return svm_predict(MeanMod, &node);
}

double SVMRegression::PredictSD(int size)
{
	svm_node node;
	node.index = 1;
	node.value = size;
	return svm_predict(SdMod, &node);
}
